#include "ns8390.h"
#include <stdio.h>

#define NS8390_RXE "NS8390 RXE Receiver Error\n"
#define NS8390_TXE "NS8390 TXE Transmitter Error\n"
#define NS8390_OVW "NS8390 OVW Buffer Overflow\n"
#define NS8390_CNT "NS8390 CNT\n"

#define NS8390_NIF_NAME "ns8390"


static void set_dma_read_at( NS8390 *ns8390, int addr, int size )
{
  /* abort any current DMA transfers */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_AB; 

  /* set the size of the dma transfer */
  ns8390->regs.page0.RBCR1 = (size >> 8) & 0xFF;
  ns8390->regs.page0.RBCR0 = size & 0xFF;

  /* set the destination of the dma transfer */
  ns8390->regs.page0.CRDA1 = (addr >> 8) & 0xFF;
  ns8390->regs.page0.CRDA0 = addr & 0xFF;

  /* start remote read */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_RD;
}

static void set_dma_write_at( NS8390 *ns8390, int addr, int size )
{
  int dummy_read;

  /* perform a dummy dma read - tell me what 70 and 2 are !?! */
  set_dma_read_at( ns8390, 70, 2);
  
  dummy_read = ns8390->regs.page0.RBCR0; /* dummy read */

  /* abort any current DMA transfers */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* set the size of the dma transfer */
  ns8390->regs.page0.RBCR1 = (size >> 8) & 0xFF;
  ns8390->regs.page0.RBCR0 = size & 0xFF;

  /* set the source of the dma transfer */
  ns8390->regs.page0.CRDA1 = (addr >> 8) & 0xFF;
  ns8390->regs.page0.CRDA0 = addr & 0xFF;

  /* start remote write */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_WR;

}

void read_eth_addr( NS8390 *ns8390, HWA_ADDR_P hwa )
{
  int i;
  UINT8 eeprom[32];
  
  /* read the ethernet address from eprom ? */

  /* abort any current DMA transfers and stop the chip */
  ns8390->CR = NS8390_CR_RD_AB | NS8390_CR_STP;

  /* set FIFO Threshold to 8bytes */
  ns8390->regs.page0.CNTR1 = NS8390_DCR_FT_8B;

  /* Reset the dma transfer size */
  ns8390->regs.page0.RBCR0 = 0;
  ns8390->regs.page0.RBCR1 = 0;

  /* mask and clear all the chip's interrupts */
  ns8390->regs.page0.CNTR2 = 0;
  ns8390->regs.page0.ISR = 0xFFFF;

  /* Set to do an internal loopback *//* MON seems undocumented */
  ns8390->regs.page0.RSR = NS8390_RCR_MON;
  ns8390->regs.page0.CNTR0 = NS8390_TCR_LOOP_I0;

  /* set transfer size to 32 bytes(sizeof EEPROM) */
  ns8390->regs.page0.RBCR0 = 32;
  ns8390->regs.page0.RBCR1 = 0;

  /* make transfer go to 0 */
  ns8390->regs.page0.CRDA0 = 0;
  ns8390->regs.page0.CRDA1 = 0;

  /* start the transfer */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_RD;

  /* load in the EEPROM data */
  for (i = 0; i < 32; ++i)
    eeprom[i] = NS8390_DATAPORT( ns8390 );

  /* suck out the Ethernet address that we need */
  for (i = 0; i < 6; i++)
    hwa[i] = eeprom[i*2];

  /* Abort any transfer and stop the chip */
  ns8390->CR = NS8390_CR_STP | NS8390_CR_RD_AB;
}


int ns8390_reset ( NIF *ne_nif )
{
  NS8390 *ns8390;
  
  /* get the address of the chip */
  ns8390 = (NS8390 *)ne_nif->nic;
 
  /* Abort any transfer and stop the chip */
  ns8390->CR = NS8390_CR_STP | NS8390_CR_RD_AB;

  /* set the Data Configuration Register defaults */
  ns8390->regs.page0.CNTR1 = NS8390_DCR_BOS | NS8390_DCR_WTS |
                             NS8390_DCR_FT_8B;

  /* zero the dma transfer size */
  ns8390->regs.page0.RBCR0 = 0;
  ns8390->regs.page0.RBCR1 = 0;

  /* set the receive config to accept broadcasts */
  ns8390->regs.page0.RSR = NS8390_RCR_AB;

  /* set the transmit config to ENDEC module loopback */
  ns8390->regs.page0.CNTR0 = NS8390_TCR_LOOP_I0;

  /* set internal boundary buffer limit */
  ns8390->regs.page0.BNRY = 70;

  /* set default internal dma address */
  ns8390->regs.page0.CLDA0 = 70;
  ns8390->regs.page0.CLDA0 = 128;

  /* clear all the chip's interrupts */
  ns8390->regs.page0.ISR = 0xFF;

  /* Enable all interrupts except for packet transmit */
  ns8390->regs.page0.CNTR2 = NS8390_IMR_CNTE |
                             NS8390_IMR_OVWE |
                             NS8390_IMR_TXEE |
                             NS8390_IMR_RXEE |
                             NS8390_IMR_PRXE;


  /* make sure we're stopped and go to page1 regs */
  ns8390->CR = NS8390_CR_STP   |
               NS8390_CR_RD_AB |
               NS8390_CR_PAGE1;

  /* program supplied hardware address into the chip */
  ns8390->regs.page1.PAR0 = ne_nif->hwa[0];
  ns8390->regs.page1.PAR1 = ne_nif->hwa[1];
  ns8390->regs.page1.PAR2 = ne_nif->hwa[2];
  ns8390->regs.page1.PAR3 = ne_nif->hwa[3];
  ns8390->regs.page1.PAR4 = ne_nif->hwa[4];
  ns8390->regs.page1.PAR5 = ne_nif->hwa[5];

  /* set current page register */
  ns8390->regs.page1.CURR = 71;
  
  /* not sure what this does - needed by SEEQ(?) */
  ne_nif->next_receive = 0x4700;

  /* program the multicast address */
  ns8390->regs.page1.MAR0 = 0;
  ns8390->regs.page1.MAR1 = 0;
  ns8390->regs.page1.MAR2 = 0;
  ns8390->regs.page1.MAR3 = 0;
  ns8390->regs.page1.MAR4 = 0;
  ns8390->regs.page1.MAR5 = 0;
  ns8390->regs.page1.MAR6 = 0;
  ns8390->regs.page1.MAR7 = 0;

  /* start up the chip */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  return TRUE;
}

void ns8390_start( NIF *ne_nif )
{
  NS8390 *ns8390;

  /* get the address of the chip */
  ns8390 = (NS8390 *)ne_nif->nic;
  
  /* start up the chip aborting any current transfer */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* set the transmit config to normal operation */
  ns8390->regs.page0.CNTR0 = 0;

  /* set the data config register */
  ns8390->regs.page0.CNTR1 = NS8390_DCR_WTS |
                             NS8390_DCR_BOS |
                             NS8390_DCR_LS  |
                             NS8390_DCR_FT_8B;
}

void ns8390_stop( NIF *ne_nif )
{
  NS8390 *ns8390;
  int    loop; 

  /* get the address of the chip */
  ns8390 = (NS8390 *)ne_nif->nic;

  /* start up the chip aborting all transfers */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_RD_AB;

  /* mask all interrupt sources */
  ns8390->regs.page0.CNTR2 = 0;

  /* clear all interrupt sources */
  ns8390->regs.page0.ISR = 0xFF;

  for (loop = 0; loop < 0x100; ++loop)
    ;

  /* stop the chip aborting all transfers */
  ns8390->CR = NS8390_CR_STP | NS8390_CR_RD_AB;

}

int ns8390_send( NIF *ne_nif, HWA_ADDR_P eth_dest, HWA_ADDR_P eth_src, UINT16 eth_type, 
                  int data_length, UINT8 *eth_data )
{
  int mod_data_length;  
  int index;
  NS8390 *ns8390;
  UINT16 *p;
  
  /* get the address of the chip */
  ns8390 = (NS8390 *)ne_nif->nic;

  /* basic checks on the data */
  if ((data_length == 0)||(data_length>ETH_MAX_SIZE))
    return FALSE;

  /* upsize runt packets... */
  mod_data_length = data_length;
  if (mod_data_length < ETH_MIN_SIZE)
    mod_data_length = ETH_MIN_SIZE;

  /* make sure we transmit an even number of bytes */
  if ((mod_data_length&1) != 0)
    mod_data_length = mod_data_length + 1;

  /* start the transmit... */
  OS_ENTER_CRITICAL()

  set_dma_write_at( ns8390, 16384, mod_data_length + 14 );

  /* output the destination ethernet address */
  p = (UINT16 *)&eth_dest;
  NS8390_DATAPORT(ns8390) = *p;
  NS8390_DATAPORT(ns8390) = *(p+2);
  NS8390_DATAPORT(ns8390) = *(p+4);

  /* output the source ethernet address */
  p = (UINT16 *)&eth_src;
  NS8390_DATAPORT(ns8390) = *p;
  NS8390_DATAPORT(ns8390) = *(p+2);
  NS8390_DATAPORT(ns8390) = *(p+4);

  /* output the frame type */
  NS8390_DATAPORT(ns8390) = eth_type;

  /* output the actual data */
  p = (UINT16 *)&eth_data;
  for (index = 0; index < data_length; index = index + 2)
  {
    NS8390_DATAPORT(ns8390) = *(p+index);
  }

  if ((data_length & 1)!=0)
  {
    NS8390_DATAPORT(ns8390) = eth_data[data_length] << 8;
    data_length = data_length + 1;
  }

  index = data_length;
  while (index<ETH_MIN_SIZE)
  {
    NS8390_DATAPORT(ns8390) = 0;
    index = index + 2;
  } 

  /* set transmit start page register */
  ns8390->regs.page0.TSR = 64;

  /* set transmit byte count msb */
  ns8390->regs.page0.NCR = (mod_data_length + 14) & 0xFF;

  /* set transmit byte count lsb */
  ns8390->regs.page0.TSR = ((mod_data_length + 14) >> 8) & 0xFF;
  
  /* Thunderpackets are Go! */
  ns8390->CR = NS8390_CR_STA | NS8390_CR_TXP | NS8390_CR_RD_AB;

  OS_EXIT_CRITICAL()
  
  return TRUE;
}

void ns8390_receive( NIF *ne_nif )
{
  NS8390 *ns8390;
  int    keep;
  int    index;
  int    not_me;
  int    frame_good;
  int    frame_length;
  UINT8  receive_status;
  UINT16 np_pointer;
  UINT16 a;
  UINT16 b;
  int    receive_count;
  UINT32 current;

  ETH_ADDR eth_dest;
  ETH_ADDR eth_src;
  UINT16   eth_type;
  NBUF     *frame; 
  UINT16   *p;
  
  /* get the address of the chip */
  ns8390 = (NS8390 *)ne_nif->nic;

  /* get the stored buffer pointer ? */
  current = ne_nif->next_receive;

  set_dma_read_at( ns8390, current, 4);

  frame_good = TRUE;

  /* read two words from chip */
  a = NS8390_DATAPORT(ns8390);
  b = NS8390_DATAPORT(ns8390);

  receive_status = a & 0xFF;
  np_pointer = (a >> 8) & 0xFF;

  receive_count = ((b & 0xFF) << 8) | ((b >> 8) & 0xFF);

  /* check the status and report and abort on error */
  if (receive_status & NS8390_RSR_PRX)
  {
    if (receive_status & NS8390_RSR_CRC)
      ne_nif->f_rx_err++;

    if (receive_status & NS8390_RSR_FAE)
      ne_nif->f_rx_err++;
  
    if (receive_status & NS8390_RSR_FO)
      ne_nif->f_rx_err++;

    if (receive_status & NS8390_RSR_MPA)
      ne_nif->f_rx_err++;

    if (receive_status & NS8390_RSR_PHY)
      ne_nif->f_rx_err++;

    if (receive_status & NS8390_RSR_DIS)
      ne_nif->f_rx_err++;

    if (receive_status == 0)
      ne_nif->f_rx_err++;

    return;
  }

  /* work out where the next internal buffer is */
  ne_nif->next_receive = np_pointer << 8;

  /* tell the chip where to go next */
  if (ne_nif->next_receive > current)
  {
    if (np_pointer < 70)
    {
      ns8390->regs.page0.CLDA0 = 0x7F;
    }
    else
    {
      ns8390->regs.page0.CLDA0 = 0xFFFF + np_pointer;
    }
  }
  else
  {
    ns8390->regs.page0.CLDA0 = np_pointer;
  }

  /* this test is redundant ? */
  if (frame_good)
  {
    /* read in the ethernet addresses and type */
    set_dma_read_at( ns8390, current + 4, 14 );
    
    /* get the destination address */
    p = (UINT16 *)&eth_dest;
    *p     = NS8390_DATAPORT(ns8390);
    *(p+2) = NS8390_DATAPORT(ns8390);
    *(p+4) = NS8390_DATAPORT(ns8390);

    /* get the source address */
    p = (UINT16 *)&eth_src;
    *p     = NS8390_DATAPORT(ns8390);
    *(p+2) = NS8390_DATAPORT(ns8390);
    *(p+4) = NS8390_DATAPORT(ns8390);

    /* get the ethernet frame type address */
    eth_type = NS8390_DATAPORT(ns8390);

    /* check the frame against registered protocols */
    if (nif_proto( ne_nif, eth_type ))
      frame_good = FALSE;

    /* make sure the frame applies to me - invalid in promiscuous mode */
    not_me = FALSE;
    for ( index = 0; index < 6; index++)
      if ( ne_nif->hwa[index] != eth_dest[index] )
        not_me = TRUE;

    if (not_me)
      frame_good = FALSE;

    /* try to allocate a receive data buffer */
    if (frame_good)
    {
      frame = nbuf_allocate();
      if (frame == NULL)
        frame_good = FALSE;
    }

    /* if a dead frame the bomb out */
    if (!frame_good)
      return;

    /* tell the chip how big the packet is */
    frame_length = receive_count - 16;
    set_dma_read_at( ns8390, current + 18, frame_length );

    /* read in the packet data and store */
    p = (UINT16 *)&frame->data;
    for (index = 0; index < frame_length; index = index + 2)
    {
      *(p+index) = NS8390_DATAPORT(ns8390);
    }
    frame->length = frame_length;

    /* get the protocol handler to look at the data */
    keep = nif_proto_handler( ne_nif, eth_type, frame->data, frame_length );

    /* if we're not keeping it - dump it */
    if (!keep)
      nbuf_release( frame );

  } /* of if frame_good */

} /* ns8390_receive */


int ns8390_handler( NS8390 *ns8390, NIF *ne_nif)
{
  UINT8 isr;

  /* stop the chip aborting all transfers */
  ns8390->CR = NS8390_CR_STP | NS8390_CR_RD_AB;

  /* Store the interrupt status */
  isr = ns8390->regs.page0.ISR;

  /* abort if invalid isr value */
  if (isr == 0)
    return FALSE;

  /* OK receive interrupt */
  if (isr & NS8390_ISR_PRX)
  {
    /* reset ISR bit and receive the packet */
    ns8390->regs.page0.ISR = NS8390_ISR_PRX;
    ns8390_receive( ne_nif );
    ne_nif->f_rx++;
  }

  /* OK transmit - just count*/
  if (isr & NS8390_ISR_PTX)
  {
    ns8390->regs.page0.ISR = NS8390_ISR_PTX;
    ne_nif->f_tx++;
  }

  /* receive error */
  if (isr & NS8390_ISR_RXE)
  {
    printf(NS8390_RXE);
    ns8390->regs.page0.ISR = NS8390_ISR_RXE;
  }

  /* transmit error - count and report */
  if (isr & NS8390_ISR_TXE)
  {
    printf(NS8390_TXE);
    ns8390->regs.page0.ISR = NS8390_ISR_TXE;
    ne_nif->f_tx_err++;
  }

  /* buffer overflow error - count and report */
  if (isr & NS8390_ISR_OVW)
  {
    printf(NS8390_OVW);
    ns8390->regs.page0.ISR = NS8390_ISR_OVW;
    ne_nif->f_rx_err++;
  }

  /* counter overflow error - report */
  if (isr & NS8390_ISR_CNT)
  {
    printf(NS8390_CNT);
    ns8390->regs.page0.ISR = NS8390_ISR_CNT;
  }

  /* remote dma complete - ignore */
  if (isr & NS8390_ISR_RDC)
    ns8390->regs.page0.ISR = NS8390_ISR_RDC;

  return TRUE;
}

int ns8390_init( NIF *ne_nif, void *base, int vector )
{
  /* create the interface */
  nif_create( ne_nif, NS8390_NIF_NAME );

  /* initialise the interface */
  ne_nif->nic = base;
  ne_nif->vector = vector;

  /* find and store "my" ethernet address */
  read_eth_addr( (NS8390 *)ne_nif->nic, ne_nif->hwa );

  ne_nif->hwa_size = 6; /* ethernet address size */
  ne_nif->mtu = ETH_MAX_SIZE;

  /* store the interface procs */
  ne_nif->reset   = &ns8390_reset;
  ne_nif->start   = &ns8390_start;
  ne_nif->stop    = &ns8390_stop;
  ne_nif->send    = &ns8390_send;
  ne_nif->receive = &ns8390_receive;

  return TRUE;  
}
